from wendelin.bigarray.array_zodb import ZBigArray
import transaction
import numpy as np
import pandas as pd
from numpy import float64, dtype, cumsum, sin, uint8
from DateTime import DateTime
from time import time
from contextlib import contextmanager
import sys
import psutil
import os
import cProfile
import pstats


def bigger_than_memory_write(self, out_of_core_index=False, zblk_format='0', **kwargs):
  message_list = []
  message_list.append('Write start: ' + get_process_memory_usage())

  _, array_schema = get_field_names_and_schema()
  float_dtype = np.dtype(float64)
                
  array_memory_size = 251*MB
  array_number_of_items = array_memory_size // array_schema.itemsize
  message_list.append(set_zblock_format(zblk_format))
  
  big_array = create_zbigarray(memory_size=array_memory_size, dtype=array_schema)

  root = self.getPhysicalRoot()
  with timer('Create arrays instances', message_list):
    if out_of_core_index:
      big_index = create_zbigarray(array_number_of_items*float_dtype.itemsize, float_dtype)
      big_array, big_index = store_arrays(root, [big_array, 'big_array'], [big_index, 'big_index'])
    else:
      big_index = np.arange(array_number_of_items)
      big_array = store_arrays(root, [big_array, 'big_array'])[0]
  
  check_zblock_format(zblk_format)
  
  message_list.append('Populating array with %s rows.' % array_number_of_items)
  response = populate_array(self, big_array, big_index, array_schema)
  message_list.extend(response['messages'])
  
  for message in response['messages']:
    if 'ZBigArray' in message:
      return json.dumps({'messages': message_list, 'result': message})
  message_list.append('Write end: ' + get_process_memory_usage())
  return json.dumps(message_list)

def bigger_than_memory_read(self, out_of_core_index=False, tool='wendelin.numpy'):
  
  message_list = []
  message_list.append('Reading with format %s' % tool)
  message_list.append('Read start ' + get_process_memory_usage())
  
  root = self.getPhysicalRoot()
  
  # We don't have the columns names here, so we set manually the 14th column
  # name to 'quantity' to simplify things.
  columns = range(23)
  columns[13] = 'quantity'
  
  ob = root.big_index
  message_list.append(len(ob._p_jar.db().storage.load(ob._p_oid, '')[0])+42)
  
  number_of_items = len(root.big_array)
  item_size = root.big_array.dtype.itemsize
  message_list.append('Processing %s items with %s bytes each' % (number_of_items, item_size))
  message_list.append(len(root.big_array._p_jar.db().storage.load(root.big_array._p_oid, '')[0])+42)
  
  big_index = root.big_index[:] if out_of_core_index else np.arange(number_of_items)
  
  messages = process_data(self, root.big_array, big_index, columns, tool)
  message_list.extend(messages)
  message_list.append('Read end: ' + get_process_memory_usage())
  
  for message in messages:
    for tool in ['wendelin.pandas', 'wendelin.numpy', 'numpy.memory', 'numpy.memmap']:
      if tool in message:
        return json.dumps({'messages': message_list, 'result': message})
  return json.dumps(message_list)

@contextmanager
def timer(name, message_list):
  '''A context manager to time function calls and add the result to
  message_list.
  '''
  start_time = time()
  yield
  elapsed_time = time() - start_time
  message_list.append('[%s] finished in %s seconds' % (name, elapsed_time))

def get_process_memory_usage():
  '''Return the runnign process' memory usage.'''
  process = psutil.Process(os.getpid())
  m = process.memory_info()
  return 'VIRT: %i MB\tRSS: %iMB' % (m.vms//MB, m.rss//MB)
  
def get_available_memory():
  return psutil.virtual_memory().available

def get_field_names_and_schema():
  '''Returns the array's dtype and its field names.'''
  field_names = ['f%s' % number if not number == 13 else 'quantity' for number in range(23)]
  array_schema = np.dtype({
    'names' : field_names,
    'formats': map(
      np.dtype,
      ['i8', 'i8', 'i8', 'i8', 'i8', 'i8', 'i8', 'i8', 'i8', 'i8',
        'i8', 'i8', 'i8', 'f8', 'i8', 'i8', 'f8', 'f8', 'f8',
        'a50', 'a50', 'a50', 'a50']
    )
  })
  return field_names, array_schema

def create_zbigarray(memory_size=None, dtype=None):
  '''Create a ZBigArray with memory_size bytes with the defined dtype.'''
  element_number = memory_size // dtype.itemsize
  return ZBigArray((element_number,), dtype)

def set_zblock_format(zblk_format):
  '''Set the array to write with the provide zblock format.
  Formats are:
    - 0 (low-overhead access timed high-overhead storage size)
    - 1 (low overhead storage size and high-overhead access time)
  '''
  file_zodb.ZBlk_fmt_write = zblk_format_string = 'ZBlk%s' % zblk_format
  return 'Writting with format: %s' % zblk_format_string

def check_zblock_format(zblk_format):
  if not file_zodb.ZBlk_fmt_write == 'ZBlk%s' % zblk_format:
    raise 'Zblk format didnt match'
  else:
    return True
  
def store_arrays(root, *arrays_filenames):
  '''Store all [array, filename] pair in arrays_filenames in root and return an
  ordered list with all store arrays.
  '''
  for array, filename in arrays_filenames:
    setattr(root, filename, None)
    setattr(root, filename, array)
  transaction.commit()
  return [getattr(root, filename) for _, filename in arrays_filenames]

def populate_array(root, big_array, big_index, schema):
  '''Fill big_array with copies of a MySQL row until it is completely full'''
  message_list = []
  with timer('Time to prepare data for assignment', message_list):
    data = root.stock_offset(my_offset=0, my_limit=1)
    columns = data._names
    row = tuple([filter_item(item, normalize=True) for item in data.tuples()[0]])
  
  max_memory_usage =  100*MB
  message_list.append('%s megabytes of memory can be used per chunk.' % (max_memory_usage//MB))
  message_list.append('Total array size is %s megabytes.' % (big_array.nbytes//MB))
  # little trick here to do a ceil-division instead of floor-division: ceil(a / b) = -(-a // b)
  chunks = -(-big_array.nbytes // max_memory_usage)
  chunk_begin = 0
  chunksize = len(big_array) // chunks
  message_list.append('Using %s chunks of %s items.' % (chunks, chunksize))
      
  with timer('Complete assign & store ZBigArray', message_list):
    for i in range(chunks):
      big_array[chunk_begin:chunksize*(i+1)] = row
      chunk_begin = chunksize + 1
      transaction.commit()
  return {'messages': message_list, 'columns': columns}

def filter_item(item, normalize=False):
  '''Typecast item to numeric values if it is a: DateTime or falsy to help
  pandas/numpy deal with them. If normalize is True it will typecast strings
  and unicodes to floats too, thus reducing that size dramatically.
  '''
  if not item or isinstance(item, type(None)):
    return 0
  if normalize and isinstance(item, (str, unicode)):
    return 0
  elif isinstance(item, DateTime):
    return 0
  else:
    return item

def process_data(root, big_array, big_index, columns, tool):
  '''Process all data in big_array. Currently, does a complete sum of the
  quantity column both with wendein.core/numpy and pandas.
  '''
  message_list = []
  result = None

  if tool == 'wendelin.pandas':
    try:
      with timer(tool, message_list):
        df = pd.DataFrame.from_records(big_array[:], index=big_index)
        result = df.quantity.sum()
      if result:
        message_list.append('Pandas result: %s' % result)
    except MemoryError:
      message_list.append('MemoryError while creating pandas.Dataframe!')
  
  if tool == 'wendelin.numpy':
    with timer(tool, message_list):
      result = big_array[:]['quantity'].sum()
    message_list.append('NumPy result: %s' % result)
  
  # common setup for in memory and memmap
  if tool == 'numpy.memory' or tool == 'numpy.memmap':
    data = root.stock_offset(my_offset=0, my_limit=1)
    columns = data._names
    row = tuple([filter_item(item, normalize=True) for item in data.tuples()[0]])
    _, schema = get_field_names_and_schema()
    array = np.ndarray((1430394,), schema)
    array[:] = row
  
  if tool == 'numpy.memory':
    with timer(tool, message_list):
      result = array['quantity'].sum()
    message_list.append('numpy in memory result: %s' % result)

  if tool == 'numpy.memmap':
    # The code commented below was just ran one time to write the array
    # for NumPy's memmap tool to avoid any type of cache from Python side.
    #
    # import os.path as path
    # filename = path.join('/tmp', 'numpy.dat')
    # write_fp = np.memmap(filename, dtype=schema, mode='w+', shape=(1430394,))
    # with timer('time to write numpy memmap', message_list):
    #   write_fp[:] = row
    #   write_fp.flush()
    with timer(tool, message_list):
      read_fp = np.memmap(filename, dtype=schema, mode='r', shape=(1430394,))
      array = np.ndarray((1430394,), schema, buffer=read_fp)
      result = array['quantity'].sum()
    message_list.append('numpy memmap result: %s' % result)

  return message_list